from __future__ import absolute_import, division, print_function

from .fcn8 import *
from .fcn16 import *
from .fcn32 import *
from .unet import *
from .pspnet import *

model_from_name = {}

model_from_name["fcn8_vgg11"] = fcn8_vgg11
model_from_name["fcn8_vgg13"] = fcn8_vgg13
model_from_name["fcn8_vgg16"] = fcn8_vgg16
model_from_name["fcn8_vgg19"] = fcn8_vgg19
model_from_name["fcn16_vgg11"] = fcn16_vgg11
model_from_name["fcn16_vgg13"] = fcn16_vgg13
model_from_name["fcn16_vgg16"] = fcn16_vgg16
model_from_name["fcn16_vgg19"] = fcn16_vgg19
model_from_name["fcn32_vgg11"] = fcn32_vgg11
model_from_name["fcn32_vgg13"] = fcn32_vgg13
model_from_name["fcn32_vgg16"] = fcn32_vgg16
model_from_name["fcn32_vgg19"] = fcn32_vgg19
model_from_name["fcn8_resnet18"] = fcn8_resnet18
model_from_name["fcn8_resnet34"] = fcn8_resnet34
model_from_name["fcn8_resnet50"] = fcn8_resnet50
model_from_name["fcn8_resnet101"] = fcn8_resnet101
model_from_name["fcn8_resnet152"] = fcn8_resnet152
model_from_name["fcn16_resnet18"] = fcn16_resnet18
model_from_name["fcn16_resnet34"] = fcn16_resnet34
model_from_name["fcn16_resnet50"] = fcn16_resnet50
model_from_name["fcn16_resnet101"] = fcn16_resnet101
model_from_name["fcn16_resnet152"] = fcn16_resnet152
model_from_name["fcn32_resnet18"] = fcn32_resnet18
model_from_name["fcn32_resnet34"] = fcn32_resnet34
model_from_name["fcn32_resnet50"] = fcn32_resnet50
model_from_name["fcn32_resnet101"] = fcn32_resnet101
model_from_name["fcn32_resnet152"] = fcn32_resnet152
model_from_name["fcn8_mobilenet_v2"] = fcn8_mobilenet_v2
model_from_name["fcn16_mobilenet_v2"] = fcn16_mobilenet_v2
model_from_name["fcn32_mobilenet_v2"] = fcn32_mobilenet_v2

model_from_name["unet"] = unet
model_from_name["unet_vgg11"] = unet_vgg11
model_from_name["unet_vgg13"] = unet_vgg13
model_from_name["unet_vgg16"] = unet_vgg16
model_from_name["unet_vgg19"] = unet_vgg19
model_from_name["unet_resnet18"] = unet_resnet18
model_from_name["unet_resnet34"] = unet_resnet34
model_from_name["unet_resnet50"] = unet_resnet50
model_from_name["unet_resnet101"] = unet_resnet101
model_from_name["unet_resnet152"] = unet_resnet152
model_from_name["unet_mobilenet_v2"] = unet_mobilenet_v2

model_from_name["pspnet_vgg11"] = pspnet_vgg11
model_from_name["pspnet_vgg13"] = pspnet_vgg13
model_from_name["pspnet_vgg16"] = pspnet_vgg16
model_from_name["pspnet_vgg19"] = pspnet_vgg19
model_from_name["pspnet_resnet18"] = pspnet_resnet18
model_from_name["pspnet_resnet34"] = pspnet_resnet34
model_from_name["pspnet_resnet50"] = pspnet_resnet50
model_from_name["pspnet_resnet101"] = pspnet_resnet101
model_from_name["pspnet_resnet152"] = pspnet_resnet152
model_from_name["pspnet_mobilenet_v2"] = pspnet_mobilenet_v2







"""
The trainer class.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

from util.validation import *
from util.logger import *

try:
    from tqdm import tqdm
    from tqdm import trange
except ImportError:
    print("tqdm and trange not found, disabling progress bars")

    def tqdm(iter):
        return iter

    def trange(iter):
        return iter

TQDM_COLS = 80

def cross_entropy2d(input, target):
    # input: (n, c, h, w), target: (n, h, w)
    n, c, h, w = input.size()

    # input: (n*h*w, c)
    input = input.transpose(1, 2).transpose(2, 3).contiguous()
    input = input[target.view(n, h, w, 1).repeat(1, 1, 1, c) >= 0]
    input = input.view(-1, c)

    # target: (n*h*w,)
    mask = target >= 0.0
    target = target[mask]

    func_loss = torch.nn.CrossEntropyLoss()
    loss = func_loss(input, target)

    return loss


class Trainer(object):

    def __init__(self, model, optimizer, logger, num_epochs, train_loader,
                 test_loader=None,
                 epoch=0,
                 log_batch_stride=30,
                 check_point_epoch_stride=60,
                 scheduler=None):
        """
        :param model: A network model to train.
        :param optimizer: A optimizer.
        :param logger: The logger for writing results to Tensorboard.
        :param num_epochs: iteration count.
        :param train_loader: pytorch's DataLoader
        :param test_loader: pytorch's DataLoader
        :param epoch: the start epoch number.
        :param log_batch_stride: it determines the step to write log in the batch loop.
        :param check_point_epoch_stride: it determines the step to save a model in the epoch loop.
        :param scheduler: optimizer scheduler for adjusting learning rate.
        """
        self.cuda = torch.cuda.is_available()
        self.model = model
        self.optim = optimizer
        self.logger = logger
        self.train_loader = train_loader
        self.test_loader = test_loader
        self.num_epoches = num_epochs
        self.check_point_step = check_point_epoch_stride
        self.log_batch_stride = log_batch_stride
        self.scheduler = scheduler

        self.epoch = epoch

    def train(self):
        if not next(self.model.parameters()).is_cuda and self.cuda:
            raise ValueError("A model should be set via .cuda() before constructing optimizer.")

        for epoch in trange(self.epoch, self.num_epoches,
                          position=0,
                          desc='Train', ncols=TQDM_COLS):
            self.epoch = epoch

            # train
            self._train_epoch()

            # step forward to reduce the learning rate in the optimizer.
            if self.scheduler:
                self.scheduler.step()

            # model checkpoints
            if epoch%self.check_point_step == 0:
                self.logger.save_model_and_optimizer(self.model,
                                                     self.optim,
                                                     'epoch_{}'.format(epoch))



    def evaluate(self):
        num_batches = len(self.test_loader)
        self.model.eval()

        with torch.no_grad():
            for n_batch, (sample_batched) in tqdm(enumerate(self.test_loader),
                                total=num_batches,
                                leave=False,
                                desc="Valid epoch={}".format(self.epoch),
                                ncols=TQDM_COLS):
                self._eval_batch(sample_batched, n_batch, num_batches)

    def _train_epoch(self):

        num_batches = len(self.train_loader)

        if self.test_loader:
            dataloader_iterator = iter(self.test_loader)

        for n_batch, (sample_batched) in tqdm(enumerate(self.train_loader),
                                              total=num_batches,
                                              leave=False,
                                              desc="Train epoch={}".format(self.epoch),
                                              ncols=TQDM_COLS):
            self.model.train()
            data = sample_batched['image']
            target = sample_batched['annotation']

            if self.cuda:
                data, target = data.cuda(), target.cuda()

            self.optim.zero_grad()

            torch.cuda.empty_cache()

            score = self.model(data)
            loss = cross_entropy2d(score, target)

            loss_data = loss.data.item()
            if np.isnan(loss_data):
                raise ValueError('loss is nan while training')

            loss.backward()
            self.optim.step()

            if n_batch%self.log_batch_stride != 0:
                continue

            self.logger.store_checkpoint_var('img_width', data.shape[3])
            self.logger.store_checkpoint_var('img_height', data.shape[2])

            self.model.img_width = data.shape[3]
            self.model.img_height = data.shape[2]

            #write logs to Tensorboard.
            lbl_pred = score.data.max(1)[1].cpu().numpy()[:, :, :]
            lbl_true = target.data.cpu().numpy()
            acc, acc_cls, mean_iou, fwavacc = \
                label_accuracy_score(lbl_true, lbl_pred, n_class=score.shape[1])

            self.logger.log_train(loss, 'loss', self.epoch, n_batch, num_batches)
            self.logger.log_train(acc, 'acc', self.epoch, n_batch, num_batches)
            self.logger.log_train(acc_cls, 'acc_cls', self.epoch, n_batch, num_batches)
            self.logger.log_train(mean_iou, 'mean_iou', self.epoch, n_batch, num_batches)
            self.logger.log_train(fwavacc, 'fwavacc', self.epoch, n_batch, num_batches)

            #write result images when starting epoch.
            if n_batch == 0:
                log_img = self.logger.concatenate_images([lbl_pred, lbl_true], input_axis='byx', normalize_uint8=True)
                log_img = self.logger.concatenate_images([log_img, data.cpu().numpy()[:, :, :, :]])
                self.logger.log_images_train(log_img, self.epoch, n_batch, num_batches,
                                             nrows=data.shape[0])

            #if the trainer has the test loader, it evaluates the model using the test data.
            if self.test_loader:
                self.model.eval()
                with torch.no_grad():
                    try:
                        sample_batched = next(dataloader_iterator)
                    except StopIteration:
                        dataloader_iterator = iter(self.test_loader)
                        sample_batched = next(dataloader_iterator)

                    self._eval_batch(sample_batched, n_batch, num_batches)


    def _eval_batch(self, sample_batched, n_batch, num_batches):
        data = sample_batched['image']
        target = sample_batched['annotation']

        if self.cuda:
            data, target = data.cuda(), target.cuda()
        torch.cuda.empty_cache()

        score = self.model(data)

        loss = cross_entropy2d(score, target)
        loss_data = loss.data.item()
        if np.isnan(loss_data):
            raise ValueError('loss is nan while training')

        lbl_pred = score.data.max(1)[1].cpu().numpy()[:, :, :]
        lbl_true = target.data.cpu().numpy()
        acc, acc_cls, mean_iou, fwavacc = \
            label_accuracy_score(lbl_true, lbl_pred, n_class=score.shape[1])

        self.logger.log_test(loss, 'loss', self.epoch, n_batch, num_batches)
        self.logger.log_test(acc, 'acc', self.epoch, n_batch, num_batches)
        self.logger.log_test(acc_cls, 'acc_cls', self.epoch, n_batch, num_batches)
        self.logger.log_test(mean_iou, 'mean_iou', self.epoch, n_batch, num_batches)
        self.logger.log_test(fwavacc, 'fwavacc', self.epoch, n_batch, num_batches)

        if n_batch == 0:
            log_img = self.logger.concatenate_images([lbl_pred, lbl_true], input_axis='byx', normalize_uint8=True)
            log_img = self.logger.concatenate_images([log_img, data.cpu().numpy()[:, :, :, :]])
            self.logger.log_images_test(log_img, self.epoch, n_batch, num_batches,
                                        nrows=data.shape[0])

    def _write_img(self, score, target, input_img, n_batch, num_batches):
        lbl_pred = score.data.max(1)[1].cpu().numpy()[:, :, :]
        lbl_true = target.data.cpu().numpy()

        log_img = self.logger.concatenate_images([lbl_pred, lbl_true], input_axis='byx', normalize_uint8=True)
        log_img = self.logger.concatenate_images([log_img, input_img.cpu().numpy()[:, :, :, :]])
        self.logger.log_images(log_img, self.epoch, n_batch, num_batches, nrows=log_img.shape[0])


"""
Dataset class.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1, OpenCV-Python 4.1.1.26
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, print_function, division
import os
import numpy as np
import time
import torch
from torch.utils.data import Dataset
import cv2

# Ignore warnings
import warnings
warnings.filterwarnings("ignore")

class DataLoaderError(Exception):
    pass

try:
    from tqdm import tqdm
except ImportError:
    print("tqdm not found, disabling progress bars")

    def tqdm(iter):
        return iter

TQDM_COLS = 80

class SegmentationDataset(Dataset):
	""" Segmentation dataset"""
	def __init__(self, images_dir, segs_dir, n_classes, transform=None):
		"""
		input images must be matched.

		:param images_dir: path to the image directory
		:param segs_dir: path to the annotation image directory
		:param n_classes: a number of the classes
		:param transform: optional transform to be applied on an image
		"""
		super(SegmentationDataset, self).__init__()

		self.images_dir = images_dir
		self.segs_dir = segs_dir
		self.transform = transform
		self.n_classes = n_classes

		self.pairs_dir = self._get_image_pairs_(self.images_dir, self.segs_dir)
		verified = self._verify_segmentation_dataset()
		assert verified

	def __len__(self):
		return len(self.pairs_dir)

	def __getitem__(self, idx):
		if torch.is_tensor(idx):
			idx = idx.tolist()

		im = cv2.imread(self.pairs_dir[idx][0], flags=cv2.IMREAD_COLOR)
		lbl = cv2.imread(self.pairs_dir[idx][1], flags=cv2.IMREAD_GRAYSCALE)

		sample = {'image': im, 'labeled': lbl}

		if self.transform:
			sample = self.transform(sample)

		return sample
	def _verify_segmentation_dataset(self):
		try:
			if not len(self.pairs_dir):
				print("Couldn't load any data from self.images_dir: "
					  "{0} and segmentations path: {1}"
					  .format(self.images_dir, self.segs_dir))
				return False

			return_value = True
			for im_fn, seg_fn in tqdm(self.pairs_dir, ncols=TQDM_COLS):
				img = cv2.imread(im_fn)
				seg = cv2.imread(seg_fn)
				# Check dimensions match
				if not img.shape == seg.shape:
					return_value = False
					print("The size of image {0} and its segmentation {1} "
						  "doesn't match (possibly the files are corrupt)."
						  .format(im_fn, seg_fn))
				else:
					max_pixel_value = np.max(seg[:, :, 0])
					if max_pixel_value >= self.n_classes:
						return_value = False
						print("The pixel values of the segmentation image {0} "
							  "violating range [0, {1}]. "
							  "Found maximum pixel value {2}"
							  .format(seg_fn, str(self.n_classes - 1), max_pixel_value))

			time.sleep(0.0001)
			if return_value:
				print("Dataset verified! ")
			else:
				print("Dataset not verified!")
			return return_value
		except DataLoaderError as e:
			print("Found error during data loading\n{0}".format(str(e)))
			return False

	def _get_image_pairs_(self, img_path1, img_path2):
		""" Check two images have the same name and get all the images
		:param img_path1: directory
		:param img_path2: directory
		:return: pair paths
		"""

		AVAILABLE_IMAGE_FORMATS = [".jpg", ".jpeg", ".png", ".bmp"]

		files1 = []
		files2 = {}
		
		for dir_entry in os.listdir(img_path1):
			if os.path.isfile(os.path.join(img_path1, dir_entry)) and \
					os.path.splitext(dir_entry)[1] in AVAILABLE_IMAGE_FORMATS:
				file_name, file_extension = os.path.splitext(dir_entry)
				files1.append((file_name, file_extension,
									os.path.join(img_path1, dir_entry)))

		for dir_entry in os.listdir(img_path2):
			if os.path.isfile(os.path.join(img_path2, dir_entry)) and \
				os.path.splitext(dir_entry)[1] in AVAILABLE_IMAGE_FORMATS:
				file_name, file_extension = os.path.splitext(dir_entry)
				full_dir_entry = os.path.join(img_path2, dir_entry)
				if file_name in files2:
					raise DataLoaderError("img_path2 with filename {0}"
										  " already exists and is ambiguous to"
										  " resolve with path {1}."
										  " Please remove or rename the latter."
										  .format(file_name, full_dir_entry))

				files2[file_name] = (file_extension, full_dir_entry)

		return_value = []
		# Match two paths
		for image_file, _, image_full_path in files1:
			if image_file in files2:
				return_value.append((image_full_path,
									 files2[image_file][1]))
			else:
				# Error out
				raise DataLoaderError("No corresponding images "
									  "found for image {0}."
									  .format(image_full_path))

		return return_value
			





from torchvision import transforms

from segmentation.data_loader.segmentation_dataset import SegmentationDataset
from segmentation.data_loader.transform import Rescale, ToTensor
from segmentation.trainer import Trainer
from segmentation.predict import *
from segmentation.models import all_models
from util.logger import Logger

train_images = r'dataset/cityspaces/images/train'
test_images = r'dataset/cityspaces/images/test'
train_labled = r'dataset/cityspaces/labeled/train'
test_labeled = r'dataset/cityspaces/labeled/test'

if __name__ == '__main__':
    model_name = "fcn8_vgg16"
    device = 'cuda'
    batch_size = 4
    n_classes = 34
    num_epochs = 10
    image_axis_minimum_size = 200
    pretrained = True
    fixed_feature = False

    logger = Logger(model_name=model_name, data_name='example')

    ### Loader
    compose = transforms.Compose([
        Rescale(image_axis_minimum_size),
        ToTensor()
         ])

    train_datasets = SegmentationDataset(train_images, train_labled, n_classes, compose)
    train_loader = torch.utils.data.DataLoader(train_datasets, batch_size=batch_size, shuffle=True, drop_last=True)

    test_datasets = SegmentationDataset(test_images, test_labeled, n_classes, compose)
    test_loader = torch.utils.data.DataLoader(test_datasets, batch_size=batch_size, shuffle=True, drop_last=True)

    ### Model
    model = all_models.model_from_name[model_name](n_classes, batch_size,
                                                   pretrained=pretrained,
                                                   fixed_feature=fixed_feature)
    model.to(device)

    ###Load model
    ###please check the foloder: (.segmentation/test/runs/models)
    #logger.load_model(model, 'epoch_15')

    ### Optimizers
    if pretrained and fixed_feature: #fine tunning
        params_to_update = model.parameters()
        print("Params to learn:")
        params_to_update = []
        for name, param in model.named_parameters():
            if param.requires_grad == True:
                params_to_update.append(param)
                print("\t", name)
        optimizer = torch.optim.Adadelta(params_to_update)
    else:
        optimizer = torch.optim.Adadelta(model.parameters())

    ### Train
    #scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=10, gamma=0.1)
    trainer = Trainer(model, optimizer, logger, num_epochs, train_loader, test_loader)
    trainer.train()


    #### Writing the predict result.
    predict(model, r'dataset/cityspaces/input.png',
             r'dataset/cityspaces/output.png')





"""
The predict functions.
The main function is to write output on the color from the gray labeled image.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1, OpenCV-Python 4.1.1.26
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import random
import cv2
import torch
import numpy as np
import os
import pathlib
import six

def parent(path):
	path = pathlib.Path(path)
	return str(path.parent)

def exist(path):
	return os.path.exists(str(path))

def mkdir(path):
	pathlib.Path(path).mkdir(parents=True, exist_ok=True)

random.seed(0)
class_colors = [(random.randint(0, 255), random.randint(
    0, 255), random.randint(0, 255)) for _ in range(5000)]

def convert_seg_gray_to_color(input, n_classes, output_path=None, colors=class_colors):
	"""
	Convert the segmented image on gray to color.

	:param input: it is available to get two type(ndarray, string), string type is a file path.
	:param n_classes: number of the classes.
	:param output_path: output path. if it is None, this function return result array(ndarray)
	:param colors: refer to 'class_colors' format. Default: random assigned color.
	:return: if out_path is None, return result array(ndarray)
	"""
	if isinstance(input, six.string_types):
		seg = cv2.imread(input, flags=cv2.IMREAD_GRAYSCALE)
	elif type(input) is np.ndarray:
		assert len(input.shape) == 2, "Input should be h,w "
		seg = input

	height = seg.shape[0]
	width = seg.shape[1]

	seg_img = np.zeros((height, width, 3))

	for c in range(n_classes):
		seg_arr = seg[:, :] == c
		seg_img[:, :, 0] += ((seg_arr) * colors[c][0]).astype('uint8')
		seg_img[:, :, 1] += ((seg_arr) * colors[c][1]).astype('uint8')
		seg_img[:, :, 2] += ((seg_arr) * colors[c][2]).astype('uint8')

	if output_path:
		cv2.imwrite(output_path, seg_img)
	else:
		return seg_img

def predict(model, input_path, output_path, colors=class_colors):
	"""
	This function can save a predicted result on the color from the trained model.

	:param model: a network model.
	:param input_path: the input file path.
	:param output_path: the output file path.
	:param colors: refer to 'class_colors' format. Default: random assigned color.
	:return: model result.
	"""
	model.eval()

	img = cv2.imread(input_path, flags=cv2.IMREAD_COLOR)
	ori_height = img.shape[0]
	ori_width = img.shape[1]

	model_width = model.img_width
	model_height = model.img_height

	if model_width != ori_width or model_height != ori_height:
		img = cv2.resize(img, (model_width, model_height), interpolation=cv2.INTER_NEAREST)


	data = img.transpose((2, 0, 1))
	data = data[None, :, :, :]
	data = torch.from_numpy(data).float()

	if next(model.parameters()).is_cuda:
		if not torch.cuda.is_available():
			raise ValueError("A model was trained via .cuda(), but this system can not support cuda.")
		data = data.cuda()

	score = model(data)

	lbl_pred = score.data.max(1)[1].cpu().numpy()[:, :, :]
	lbl_pred = lbl_pred.transpose((1, 2, 0))
	n_classes = np.max(lbl_pred)
	lbl_pred = lbl_pred.reshape(model_height, model_width)

	seg_img = convert_seg_gray_to_color(lbl_pred, n_classes, colors=colors)

	if model_width != ori_width or model_height != ori_height:
		seg_img = cv2.resize(seg_img, (ori_width, ori_height), interpolation=cv2.INTER_NEAREST)

	if not exist(parent(output_path)):
		mkdir(parent(output_path))

	cv2.imwrite(output_path, seg_img)

	return score


"""
PSPnet class.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import torch
import torch.nn.functional as F
from ..encoders.squeeze_extractor import *

class PSPModule(nn.Module):
    def __init__(self, in_channels, out_channels=1024, pool_factors=(1, 2, 3, 6), batch_norm=True):
        super().__init__()
        self.spatial_blocks = []
        for pf in pool_factors:
            self.spatial_blocks += [self._make_spatial_block(in_channels, pf, batch_norm)]
        self.spatial_blocks = nn.ModuleList(self.spatial_blocks)

        bottleneck = []
        bottleneck += [nn.Conv2d(in_channels * (len(pool_factors) + 1), out_channels, kernel_size=1)]
        if batch_norm:
            bottleneck += [nn.BatchNorm2d(out_channels)]
        bottleneck += [nn.ReLU(inplace=True)]
        self.bottleneck = nn.Sequential(*bottleneck)

    def _make_spatial_block(self, in_channels, pool_factor, batch_norm):
        spatial_block = []
        spatial_block += [nn.AdaptiveAvgPool2d(output_size=(pool_factor, pool_factor))]
        spatial_block += [nn.Conv2d(in_channels, in_channels, kernel_size=1, bias=False)]
        if batch_norm:
            spatial_block += [nn.BatchNorm2d(in_channels)]
        spatial_block += [nn.ReLU(inplace=True)]

        return nn.Sequential(*spatial_block)

    def forward(self, x):
        h, w = x.size(2), x.size(3)
        pool_outs = [x]
        for block in self.spatial_blocks:
            pooled = block(x)
            pool_outs += [F.upsample(pooled, size=(h, w), mode='bilinear')]
        o = torch.cat(pool_outs, dim=1)
        o = self.bottleneck(o)
        return o

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

class PSPUpsampling(nn.Module):
    def __init__(self, in_channels, out_channels, batch_norm=True):
        super().__init__()
        layers = []
        layers += [nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)]
        if batch_norm:
            layers += [nn.BatchNorm2d(out_channels)]
        layers += [nn.ReLU(inplace=True)]
        self.layer = nn.Sequential(*layers)

    def forward(self, x):
        h, w = 2 * x.size(2), 2 * x.size(3)
        p = F.upsample(x, size=(h, w), mode='bilinear')
        return self.layer(p)

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

class PSPnet(torch.nn.Module):

    def __init__(self, n_classes, pretrained_model: SqueezeExtractor, batch_norm=True, psp_out_feature=1024):
        super(PSPnet, self).__init__()
        self.features = pretrained_model.features

        # find out_channels of the top layer and define classifier
        for idx, m in reversed(list(enumerate(self.features.modules()))):
            if isinstance(m, nn.Conv2d):
                channels = m.out_channels
                break

        self.PSP = PSPModule(channels, out_channels=psp_out_feature, batch_norm=batch_norm)
        h_psp_out_feature = int(psp_out_feature / 2)
        q_psp_out_feature = int(psp_out_feature / 4)
        e_psp_out_feature = int(psp_out_feature / 8)
        self.upsampling1 = PSPUpsampling(psp_out_feature, h_psp_out_feature, batch_norm=batch_norm)
        self.upsampling2 = PSPUpsampling(h_psp_out_feature, q_psp_out_feature, batch_norm=batch_norm)
        self.upsampling3 = PSPUpsampling(q_psp_out_feature,  e_psp_out_feature, batch_norm=batch_norm)

        self.classifier = nn.Sequential(nn.Conv2d(e_psp_out_feature, n_classes, kernel_size=1))

        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        o = x
        for f in self.features:
            o = f(o)
            
        o = self.PSP(o)
        o = self.upsampling1(o)
        o = self.upsampling2(o)
        o = self.upsampling3(o)

        o = F.upsample(o, size=(x.shape[2], x.shape[3]), mode='bilinear')
        o = self.classifier(o)

        return o

from ..encoders.vgg import *
from ..encoders.resnet import *
from ..encoders.mobilenet import *

def pspnet_vgg11(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_11(batch_norm, pretrained, fixed_feature)
    copy_feature_info = vgg.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index - 1
    vgg.features = vgg.features[:squeeze_feature_idx]
    return PSPnet(n_classes, vgg, batch_norm)
def pspnet_vgg13(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_13(batch_norm, pretrained, fixed_feature)
    copy_feature_info = vgg.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index - 1
    vgg.features = vgg.features[:squeeze_feature_idx]
    return PSPnet(n_classes, vgg, batch_norm)
def pspnet_vgg16(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_16(batch_norm, pretrained, fixed_feature)
    copy_feature_info = vgg.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index - 1
    vgg.features = vgg.features[:squeeze_feature_idx]
    return PSPnet(n_classes, vgg, batch_norm)
def pspnet_vgg19(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_19(batch_norm, pretrained, fixed_feature)
    copy_feature_info = vgg.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index - 1
    vgg.features = vgg.features[:squeeze_feature_idx]
    return PSPnet(n_classes, vgg, batch_norm)

def pspnet_resnet18(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet18(pretrained, fixed_feature)
    copy_feature_info = resnet.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index
    resnet.features = resnet.features[:squeeze_feature_idx]
    return PSPnet(n_classes, resnet, batch_norm)
def pspnet_resnet34(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet34(pretrained, fixed_feature)
    copy_feature_info = resnet.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index
    resnet.features = resnet.features[:squeeze_feature_idx]
    return PSPnet(n_classes, resnet, batch_norm)
def pspnet_resnet50(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet50(pretrained, fixed_feature)
    copy_feature_info = resnet.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index
    resnet.features = resnet.features[:squeeze_feature_idx]
    return PSPnet(n_classes, resnet, batch_norm)
def pspnet_resnet101(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet101(pretrained, fixed_feature)
    copy_feature_info = resnet.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index
    resnet.features = resnet.features[:squeeze_feature_idx]
    return PSPnet(n_classes, resnet, batch_norm)
def pspnet_resnet152(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet152(pretrained, fixed_feature)
    copy_feature_info = resnet.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index
    resnet.features = resnet.features[:squeeze_feature_idx]
    return PSPnet(n_classes, resnet, batch_norm)

def pspnet_mobilenet_v2(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    mobile_net = mobilenet(pretrained, fixed_feature)
    copy_feature_info = mobile_net.get_copy_feature_info()
    squeeze_feature_idx = copy_feature_info[3].index
    mobile_net.features = mobile_net.features[:squeeze_feature_idx]
    return PSPnet(n_classes, mobile_net, batch_norm)




"""
Vgg16 model Customized from Torchvision.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division
from .squeeze_extractor import *
from torch import nn


class _VGG(SqueezeExtractor):
	def __init__(self, model, features, fixed_feature=True):
		super(_VGG, self).__init__(model, features, fixed_feature)

	def get_copy_feature_info(self):

		lst_copy_feature_info = []
		for i in range(len(self.features)):
			if isinstance(self.features[i], nn.MaxPool2d):
				out_channels = self._get_last_conv2d_out_channels(self.features[:i])
				lst_copy_feature_info.append(CopyFeatureInfo(i, out_channels))
		return lst_copy_feature_info

def vgg_11(batch_norm=True, pretrained=False, fixed_feature=True):
	""" VGG 11-layer model from torchvision's vgg model.

	:param batch_norm: train model with batch normalization
	:param pretrained: if true, return a model pretrained on ImageNet
	:param fixed_feature: if true and pretrained is true, model features are fixed while training.
	"""
	if batch_norm:
		from torchvision.models.vgg import vgg11_bn
		model = vgg11_bn(pretrained)
	else:
		from torchvision.models.vgg import vgg11
		model = vgg11(pretrained)

	ff = True if pretrained and fixed_feature else False
	return _VGG(model, model.features, ff)

def vgg_13(batch_norm=True, pretrained=False, fixed_feature=True):
	""" VGG 13-layer model from torchvision's vgg model.

	:param batch_norm: train model with batch normalization
	:param pretrained: if true, return a model pretrained on ImageNet
	:param fixed_feature: if true and pretrained is true, model features are fixed while training.
	"""
	if batch_norm:
		from torchvision.models.vgg import vgg13_bn
		model = vgg13_bn(pretrained)
	else:
		from torchvision.models.vgg import vgg13
		model = vgg13(pretrained)

	ff = True if pretrained and fixed_feature else False
	return _VGG(model, model.features, ff)

def vgg_16(batch_norm=True, pretrained=False, fixed_feature=True):
	""" VGG 16-layer model from torchvision's vgg model.

	:param batch_norm: train model with batch normalization
	:param pretrained: if true, return a model pretrained on ImageNet
	:param fixed_feature: if true and pretrained is true, model features are fixed while training.
	"""
	if batch_norm:
		from torchvision.models.vgg import vgg16_bn
		model = vgg16_bn(pretrained)
	else:
		from torchvision.models.vgg import vgg16
		model = vgg16(pretrained)

	ff = True if pretrained and fixed_feature else False
	return _VGG(model, model.features, ff)

def vgg_19(batch_norm=True, pretrained=False, fixed_feature=True):
	""" VGG 19-layer model from torchvision's vgg model.

	:param batch_norm: train model with batch normalization
	:param pretrained: if true, return a model pretrained on ImageNet
	:param fixed_feature: if true and pretrained is true, model features are fixed while training.
	"""
	if batch_norm:
		from torchvision.models.vgg import vgg19_bn
		model = vgg19_bn(pretrained)
	else:
		from torchvision.models.vgg import vgg19
		model = vgg19(pretrained)

	ff = True if pretrained and fixed_feature else False
	return _VGG(model, model.features, ff)



"""
The logger class. This class write logs to Tensorboard mostly.
Also, it can save images and will be stored in '.runs' path

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import os
import numpy as np
import errno
import torchvision.utils as vutils
from tensorboardX import SummaryWriter
from matplotlib import pyplot as plt
import pathlib
import datetime
from util.checkpoint import *
import copy

class Logger:

	def __init__(self, model_name, data_name):
		self.model_name = model_name
		self.data_name = data_name

		self.comment = '{}_{}'.format(model_name, data_name)
		self.data_subdir = '{}/{}'.format(model_name, data_name)

		current_time = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_")

		train_log_dir = r'runs/' + current_time + self.comment + r'/train'
		test_log_dir = r'runs/' + current_time + self.comment + r'/test'

		self.hdl_chkpoint = CheckpointHandler()

		# TensorBoard
		self.writer_train = SummaryWriter(train_log_dir, comment=self.comment)
		self.writer_test = SummaryWriter(test_log_dir, comment=self.comment)

	def log_train(self, scalar, title, epoch, n_batch, num_batches):
		step = Logger._step(epoch, n_batch, num_batches)
		self.writer_train.add_scalar(
			'{}/{}'.format(self.comment, title), scalar, step)

	def log_test(self, scalar, title, epoch, n_batch, num_batches):
		step = Logger._step(epoch, n_batch, num_batches)
		self.writer_test.add_scalar(
			'{}/{}'.format(self.comment, title), scalar, step)

	def concatenate_images(self, *args, input_axis='bcyx', normalize_uint8=False):
		"""
		This function concatenate images and return the result.

		:param images: 			available follow
								Tensor: float32
								np: uint8, int64, float32
		:param input_axis: 		if the input_axis is 'byxc', it transpose axis to 'bcyx'.
								available (bcyx | byxc | cyx | yxc | byx)
		:param normalize_uint8:	if dtype is [np.uint8], values are divided by 255.
		:return: images
		"""
		imgs = []
		for arg in args:
			if isinstance(arg, list):
				imgs += arg
			else:
				imgs.append(arg)

		assert len(imgs) >= 1

		for i in range(len(imgs)):
			if isinstance(imgs[i], np.ndarray):
				if imgs[i].dtype == np.uint8 or imgs[i].dtype == np.int64:
					if normalize_uint8:
						imgs[i] = np.clip(imgs[i].astype(np.float32) / 255, 0.0, 1.0)
					else:
						imgs[i] = imgs[i].astype(np.float32)
				imgs[i] = torch.from_numpy(imgs[i])

			if imgs[i].dtype != torch.float32:
				imgs[i] = imgs[i].float()

			if len(imgs[i].shape) == 2:  # Tensor, yx --> 11yx
				imgs[i] = imgs[i][None, None, :, :]
			elif len(imgs[i].shape) == 3:  # Tensor, cyx --> 1cyx
				imgs[i] = imgs[i][None, :, :, :]

			# swap axis to 'bcyx'
			if input_axis == 'byxc' or input_axis == 'yxc':
				imgs[i] = imgs[i].transpose(1, 3)
				imgs[i] = imgs[i].transpose(1, 2)
			elif input_axis == 'byx':
				imgs[i] = imgs[i].transpose(0, 1)
				
			if imgs[i].shape[1] == 1:
				imgs[i] = imgs[i].repeat((1, 3, 1, 1)) / 3.0

		return torch.cat(imgs)

	def log_images_train(self, images, epoch, n_batch, num_batches, input_axis='bcyx',
				   nrows=8, padding=2, pad_value=1, normalize=True, normalize_uint8=False):
		"""
		This function writes images to Tensorboard and save the file at [./data/images/]

		:param images: 			available follow
								Tensor: float32
								np: uint8, int64, float32
		:param epoch: 			epoch.
		:param n_batch:			batch index.
		:param num_batches: 	batch counts.
		:param nrows: 			grid's rows on the image.
		:param padding: 		amount of padding.
		:param pad_value: 		padding scalar value, the range [0, 1].
		:param input_axis: 		if the input_axis is 'byxc', it transpose axis to 'bcyx', available as follow
								(bcyx | byxc | cyx | yxc | byx)
		:param normalize: 		normalize image to the range [0, 1].
		:param normalize_uint8:	if dtype is [np.uint8], values are divided by 255.
		"""

		img_name, grid, step = self._log_images(images, epoch, n_batch, num_batches,
												input_axis, nrows, padding, pad_value, normalize,
						 						comment='train')

		# Add images to tensorboard
		self.writer_train.add_image(img_name, grid, step)

	def log_images_test(self, images, epoch, n_batch, num_batches, input_axis='bcyx',
				   nrows=8, padding=2, pad_value=1, normalize=True, normalize_uint8=False):
		"""
		This function writes images to Tensorboard and save the file at [./data/images/]

		:param images: 			available follow
								Tensor: float32
								np: uint8, int64, float32
		:param epoch: 			epoch.
		:param n_batch:			batch index.
		:param num_batches: 	batch counts.
		:param nrows: 			grid's rows on the image.
		:param padding: 		amount of padding.
		:param pad_value: 		padding scalar value, the range [0, 1].
		:param input_axis: 		if the input_axis is 'byxc', it transpose axis to 'bcyx', available as follow
								(bcyx | byxc | cyx | yxc | byx)
		:param normalize: 		normalize image to the range [0, 1].
		:param normalize_uint8:	if dtype is [np.uint8], values are divided by 255.
		"""

		img_name, grid, step = self._log_images(images, epoch, n_batch, num_batches,
												input_axis, nrows, padding, pad_value, normalize,
						 						comment='test')

		# Add images to tensorboard
		self.writer_test.add_image(img_name, grid, step)

	def _log_images(self, images, epoch, n_batch, num_batches, input_axis='bcyx',
				   nrows=8, padding=2, pad_value=1, normalize=True, normalize_uint8=False, comment=''):

		if isinstance(images, np.ndarray):
			if images.dtype == np.uint8 or images.dtype == np.int64:
				if normalize_uint8:
					images = np.clip(images.astype(np.float32) / 255, 0.0, 1.0)
				else:
					images = images.astype(np.float32)
			images = torch.from_numpy(images)

		if len(images.shape) == 2:  # Tensor, yx --> 11yx
			images = images[None, None, :, :]
		elif len(images.shape) == 3:  # Tensor, cyx --> 1cyx
			images = images[None, :, :, :]

		# swap axis to 'bcyx'
		if input_axis == 'byxc' or input_axis == 'yxc':
			images = images.transpose(1, 3)
			images = images.transpose(1, 2)
		elif input_axis == 'byx':
			images = images.transpose(0, 1)

		step = Logger._step(epoch, n_batch, num_batches)
		img_name = '{}/images{}'.format(self.comment, '')

		# Make grid from image tensor
		if images.shape[0] < nrows:
			nrows = images.shape[0]

		grid = vutils.make_grid(images, nrow=nrows, normalize=normalize,
								scale_each=True, pad_value=pad_value, padding=padding)

		# Save plots
		self._save_torch_images(grid, epoch, n_batch, comment)

		return img_name, grid, step

	def _save_torch_images(self, grid, epoch, n_batch, comment=''):
		out_dir = './runs/images/{}'.format(self.data_subdir)
		Logger._make_dir(out_dir)

		# Save squared
		fig = plt.figure()
		plt.imshow(np.moveaxis(grid.numpy(), 0, -1))
		plt.axis('off')
		if comment:
			fig.savefig('{}/{}_epoch_{}_batch_{}.png'.format(out_dir, comment, epoch, n_batch))
		else:
			fig.savefig('{}/epoch_{}_batch_{}.png'.format(out_dir, comment, epoch, n_batch))
		plt.close()

	def store_checkpoint_var(self, key, value):
		self.hdl_chkpoint.store_var(key, value)

	def save_model(self, model, file_name):
		out_dir = './runs/models/{}'.format(self.data_subdir)
		if not Logger._exist(out_dir):
			Logger._make_dir(out_dir)

		self.hdl_chkpoint.save_checkpoint('{}/{}'.format(out_dir, file_name))

	def save_model_and_optimizer(self, model, optim, file_name):
		out_dir = './runs/models/{}'.format(self.data_subdir)
		if not Logger._exist(out_dir):
			Logger._make_dir(out_dir)

		self.hdl_chkpoint.save_checkpoint('{}/{}'.format(out_dir, file_name), model, optim)

	def load_model(self, model, file_name):
		dir = './runs/models/{}'.format(self.data_subdir)
		assert Logger._exist(dir)

		self.hdl_chkpoint = self.hdl_chkpoint.load_checkpoint('{}/{}'.format(dir, file_name))

		model.load_state_dict(self.hdl_chkpoint.model_state_dict)
		if hasattr(self.hdl_chkpoint, '__dict__'):
			for k in self.hdl_chkpoint.__dict__:
				if k == 'model_state_dict' or k == 'optimizer_state_dict':
					continue
				attr_copy = copy.deepcopy(getattr(self.hdl_chkpoint, k))
				setattr(model, k, attr_copy)

	# def load_model_and_optimizer(self, model, optim, file_name):
	# 	dir = './runs/models/{}'.format(self.data_subdir)
	# 	assert Logger._exist(dir)
	#
	# 	self.hdl_chkpoint = self.hdl_chkpoint.load_checkpoint('{}/{}'.format(dir, file_name))
	#
	# 	model.load_state_dict(self.hdl_chkpoint.model_state_dict)
	# 	optim.load_state_dict(self.hdl_chkpoint.optimizer_state_dict)
	# 	if hasattr(self.hdl_chkpoint, '__dict__'):
	# 		for k in self.hdl_chkpoint.__dict__:
	# 			if k == 'model_state_dict' or k == 'optimizer_state_dict':
	# 				continue
	# 			attr_copy = copy.deepcopy(getattr(self.hdl_chkpoint, k))
	# 			setattr(model, k, attr_copy)

	def close(self):
		self.writer.close()

	@staticmethod
	def _step(epoch, n_batch, num_batches):
		return epoch * num_batches + n_batch

	@staticmethod
	def _make_dir(directory):
		try:
			os.makedirs(directory)
		except OSError as e:
			if e.errno != errno.EEXIST:
				raise

	@staticmethod
	def _parent(path):
		path = pathlib.Path(path)
		return str(path.parent)

	@staticmethod
	def _exist(path):
		return os.path.exists(str(path))



"""
FCN16 class.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import torch
from ..encoders.squeeze_extractor import *

class FCN16(torch.nn.Module):

	def __init__(self, n_classes, pretrained_model: SqueezeExtractor):
		super(FCN16, self).__init__()
		self.pretrained_model = pretrained_model
		self.features = self.pretrained_model.features
		self.copy_feature_info = pretrained_model.get_copy_feature_info()

		self.score_pool4 = nn.Conv2d(self.copy_feature_info[-2].out_channels,
									 n_classes, kernel_size=1)

		self.upsampling2 = nn.ConvTranspose2d(n_classes, n_classes, kernel_size=4,
											  stride=2, bias=False)
		self.upsampling16 = nn.ConvTranspose2d(n_classes, n_classes, kernel_size=32,
											   stride=16, bias=False)

		for m in self.features.modules():
			if isinstance(m, nn.Conv2d):
				channels = m.out_channels

		self.classifier = nn.Sequential(nn.Conv2d(channels, n_classes, kernel_size=1), nn.Sigmoid())
		self._initialize_weights()

	def _initialize_weights(self):
		for m in self.modules():
			if isinstance(m, nn.Conv2d):
				nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
				if m.bias is not None:
					nn.init.constant_(m.bias, 0)
			elif isinstance(m, nn.Linear):
				nn.init.normal_(m.weight, 0, 0.01)
				nn.init.constant_(m.bias, 0)

	def forward(self, x):
		last_feature_index = self.copy_feature_info[-2].index

		o = x
		for i in range(len(self.features)):
			o = self.features[i](o)
			if i == last_feature_index:
				pool4 = o

		o = self.classifier(o)
		o = self.upsampling2(o)

		o2 = self.score_pool4(pool4)
		o = o[:, :, 1:1 + o2.size()[2], 1:1 + o2.size()[3]]
		o = o + o2

		o = self.upsampling16(o)
		cx = int((o.shape[3] - x.shape[3]) / 2)
		cy = int((o.shape[2] - x.shape[2]) / 2)
		o = o[:, :, cy:cy + x.shape[2], cx:cx + x.shape[3]]

		return o

from ..encoders.vgg import *
from ..encoders.resnet import *
from ..encoders.mobilenet import *

def fcn16_vgg11(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	vgg = vgg_11(batch_norm, pretrained, fixed_feature)
	return FCN16(n_classes, vgg)
def fcn16_vgg13(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	vgg = vgg_13(batch_norm, pretrained, fixed_feature)
	return FCN16(n_classes, vgg)
def fcn16_vgg16(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	vgg = vgg_16(batch_norm, pretrained, fixed_feature)
	return FCN16(n_classes, vgg)
def fcn16_vgg19(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	vgg = vgg_19(batch_norm, pretrained, fixed_feature)
	return FCN16(n_classes, vgg)

def fcn16_resnet18(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	resnet = resnet18(pretrained, fixed_feature)
	return FCN16(n_classes, resnet)
def fcn16_resnet34(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	resnet = resnet34(pretrained, fixed_feature)
	return FCN16(n_classes, resnet)
def fcn16_resnet50(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	resnet = resnet50(pretrained, fixed_feature)
	return FCN16(n_classes, resnet)
def fcn16_resnet101(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	resnet = resnet101(pretrained, fixed_feature)
	return FCN16(n_classes, resnet)
def fcn16_resnet152(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	resnet = resnet152(pretrained, fixed_feature)
	return FCN16(n_classes, resnet)

def fcn16_mobilenet_v2(n_classes, batch_size, pretrained=False, fixed_feature=True):
	batch_norm = False if batch_size == 1 else True
	mobile_net = mobilenet(pretrained, fixed_feature)
	return FCN16(n_classes, mobile_net)


"""
imshow function. This is useful to see an tiled image.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1, OpenCV-Python 4.1.1.26, PIL
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import numpy as np
import torch
import os
if os.system('python -c "import matplotlib.pyplot as plt;plt.figure()"') != 0: # non gui support
	print("non gui system. use Agg instead")
	import matplotlib # https://stackoverflow.com/questions/43003758/
	matplotlib.use("Agg") # matplotlib-is-throwing-segmentation-fault-when-running-on-non-gui-machineweb-se
import matplotlib.pyplot as plt
from PIL import Image
import pathlib

def parent(path):
	path = pathlib.Path(path)
	return str(path.parent)

def exist(path):
	return os.path.exists(str(path))

def mkdir(path):
	pathlib.Path(path).mkdir(parents=True, exist_ok=True)

def normalize_img(img, vmin=None, vmax=None):
	"""
	:param img:	Tensor, np
	:param vmin:
	:param vmax:
	:return:		Tensor, np, float32, return the same dimension
	"""
	if isinstance(img, np.ndarray):
		img = img.astype(np.float32)
		if vmin is None:
			vmin = img.min()
		if vmax is None:
			vmax = img.max()
		img = np.clip(img, vmin, vmax)
		img = (img - vmin) / (vmax - vmin)
		img = np.clip(img, 0, 1)  # numeric error 대비
		return img
	elif isinstance(img, torch.Tensor):
		img = img.type(torch.float32)
		if vmin is None:
			vmin = img.min()
		if vmax is None:
			vmax = img.max()
		img = torch.clamp(img, vmin, vmax)
		img = (img - vmin) / (vmax - vmin)
		img = torch.clamp(img, 0, 1)
		return img
	else:
		raise ValueError

def imshow(*args, nx=None, vmin=None, vmax=None, path=None, is_color=False, normalize_uint8=False, title=None):
	"""
		:param args: available as follow
			Tensor:
				float32, bcyx | cyx
			np:
				uint8, byxc | yxc
				float32, byxc | yxc
			PIL image
		:param nx:					image counts on cols.
									(default) if counts bigger than ten, nx is fixed ten.
		:param vmin:				(default) minimum value of the image.
		:param vmax:				(default) maximum value of the image.
		:param path:				saved path. if it is null, not save.
		:param is_color:			gray or color.
		:param normalize_uint8:	    if it is [np.uint8], this param decides to divide 255.
		:param title:	    		firgure name
	"""
	# args is turned into one list.

	imgs = []
	for arg in args:
		if isinstance(arg, list):
			imgs += arg
		else:
			imgs.append(arg)

	assert len(imgs) >= 1

	for i in range(len(imgs)):
		if isinstance(imgs[i], np.ndarray):
			# if normalize_uint8 is true, values is normalized [0,1]
			if normalize_uint8 and imgs[i].dtype == np.uint8:
				imgs[i] = np.clip(imgs[i].astype(np.float32) / 255, 0, 1)
			else:
				imgs[i] = imgs[i].astype(np.float32)
			if len(imgs[i].shape) == 2:  # np, yx --> 1yx1
				imgs[i] = imgs[i][None, :, :, None]
			elif len(imgs[i].shape) == 3:  # np, yxc --> 1yxc
				imgs[i] = imgs[i][None, :, :, :]
			elif len(imgs[i].shape) == 4:
				pass
			else:
				raise ValueError
		elif isinstance(imgs[i], torch.Tensor):
			imgs[i] = imgs[i].cpu().detach().numpy()
			imgs[i] = imgs[i].astype(np.float32)
			if len(imgs[i].shape) == 2:  # Tensor, yx --> 1yx1
				imgs[i] = imgs[i][None, :, :, None]
			elif len(imgs[i].shape) == 3:  # Tensor, cyx --> 1yxc
				imgs[i] = imgs[i][None, :, :, :]
				imgs[i] = np.transpose(imgs[i], [0, 2, 3, 1])
			elif len(imgs[i].shape) == 4:
				imgs[i] = np.transpose(imgs[i], [0, 2, 3, 1])
			else:
				raise ValueError
		elif isinstance(imgs[i], Image.Image):
			imgs[i] = np.array(imgs[i]).astype(np.float32)
			imgs[i] = np.clip(imgs[i] / 255, 0, 1)
			if len(imgs[i].shape) == 2:  # PIL img, yx --> 1yx1
				imgs[i] = imgs[i][None, :, :, None]
			elif len(imgs[i].shape) == 3:  # PIL img, yxc --> 1yxc
				imgs[i] = imgs[i][None, :, :, :3]  # if it has alpha, it is trimmed.
			else:
				raise ValueError
		else:
			raise ValueError

	# imgs's byxc must be matched
	img = np.concatenate(imgs)

	# check color
	b, y, x, c = img.shape
	if is_color and c != 3 and c != 4:
		raise ValueError

	# np, bcyx, float32
	img = np.transpose(img, [0, 3, 1, 2])

	# set nx automatically from image counts
	num_img = img.shape[0] if is_color else img.shape[0] * img.shape[1]
	if nx is None:
		if num_img < 10:
			nx = num_img
		else:
			nx = 10

	if not is_color:  # gray
		ny = int(np.ceil(np.float32(b * c) / nx))
		img = img.reshape(b * c, y, x)
		black = np.zeros([ny * nx - b * c, y, x], np.float32)
		img = np.concatenate([img, black])  # ny*nx,y,x
		img = img.reshape(ny, nx, y, x)
		img = img.transpose(0, 2, 1, 3)  # ny,y,nx,x
		img = img.reshape(ny * y, nx * x)  # ny*y,nx*x complete
	else:  # color: if img is color image, it'll create three channels of the gray image.
		ny = int(np.ceil(np.float32(b) / nx))
		black = np.zeros([ny * nx - b, c, y, x], np.float32)
		img = np.concatenate([img, black])  # ny*nx,c,y,x
		img = img.reshape(ny, nx, c, y, x)
		img = img.transpose(0, 3, 1, 4, 2)  # ny,y,nx,x,c
		img = img.reshape(ny* y, nx * x, c)  # ny*y,nx*x,c complete

	# rescale
	img = normalize_img(img, vmin, vmax)

	# plot or save:
	if path == None:
		plt.figure(num=title)
		if is_color:
			fig = plt.imshow(img, interpolation="nearest", vmin=0, vmax=1)
		else:
			fig = plt.imshow(img, cmap="gray", interpolation="nearest", vmin=0, vmax=1)
		fig.axes.get_xaxis().set_visible(False)
		fig.axes.get_yaxis().set_visible(False)
		plt.pause(0.001)
		plt.show(block=False)
		plt.tight_layout()
	else:
		if not exist(parent(path)):
			mkdir(parent(path))
		img = np.clip(img * 255, 0, 255).astype(np.uint8)
		img = Image.fromarray(img)
		img.save(path)


def imshowc(*args, nx=None, vmin=0, vmax=1, path=None, normalize_uint8=True, title=None):
	"""
		Drawing a color image.

		:param args: available as follow
			Tensor:
				float32, bcyx | cyx
			np:
				uint8, bcyx | cyx
				float32, bcyx | cyx
			PIL image
		:param nx:					image counts on cols.
									(default) if counts bigger than ten, nx is fixed ten.
		:param vmin:				(default) minimum value of the image.
		:param vmax:				(default) maximum value of the image.
		:param path:				saved path. if it is null, not save.
		:param is_color:			gray or color.
		:param normalize_uint8:	    if it is [np.uint8], this param decides to divide 255.
		:param title:	    		firgure name
	"""
	imshow(*args, nx=nx, vmin=vmin, vmax=vmax, path=path, is_color=True, normalize_uint8=normalize_uint8, title=title)



"""
Resnet model Customized from Torchvision.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division
from .squeeze_extractor import *


class _ResNet(SqueezeExtractor):
    def __init__(self, model, fixed_feature=True):
        features = nn.Sequential(
            model.conv1,
            model.bn1,
            model.relu,
            model.maxpool,
            model.layer1,
            model.layer2,
            model.layer3,
            model.layer4
        )
        super(_ResNet, self).__init__(model, features, fixed_feature)

    def get_copy_feature_info(self):
        lst_copy_feature_info = []
        channel = 0
        for i in range(len(self.features)):
            feature = self.features[i]
            if isinstance(feature, nn.MaxPool2d):
                lst_copy_feature_info.append(CopyFeatureInfo(i, channel))
            for idx, m in enumerate(feature.modules()):
                if isinstance(m, nn.Conv2d) and m.stride == (2, 2):
                    channel = self._get_last_conv2d_out_channels(feature)
                    lst_copy_feature_info.append(CopyFeatureInfo(i, channel))
                    break

        return lst_copy_feature_info



def resnet18(pretrained=False, fixed_feature=True):
    """ "ResNet-18 model from torchvision's resnet model.

    :param pretrained: if true, return a model pretrained on ImageNet
    :param fixed_feature: if true and pretrained is true, model features are fixed while training.
    """
    from torchvision.models.resnet import resnet18
    model = resnet18(pretrained)

    ff = True if pretrained and fixed_feature else False
    return _ResNet(model, ff)

def resnet34(pretrained=False, fixed_feature=True):
    """ "ResNet-34 model from torchvision's resnet model.

    :param pretrained: if true, return a model pretrained on ImageNet
    :param fixed_feature: if true and pretrained is true, model features are fixed while training.
    """
    from torchvision.models.resnet import resnet34
    model = resnet34(pretrained)

    ff = True if pretrained and fixed_feature else False
    return _ResNet(model, ff)

def resnet50(pretrained=False, fixed_feature=True):
    """ "ResNet-50 model from torchvision's resnet model.

    :param pretrained: if true, return a model pretrained on ImageNet
    :param fixed_feature: if true and pretrained is true, model features are fixed while training.
    """
    from torchvision.models.resnet import resnet50
    model = resnet50(pretrained)

    ff = True if pretrained and fixed_feature else False
    return _ResNet(model, ff)

def resnet101(pretrained=False, fixed_feature=True):
    """ "ResNet-101 model from torchvision's resnet model.

    :param pretrained: if true, return a model pretrained on ImageNet
    :param fixed_feature: if true and pretrained is true, model features are fixed while training.
    """
    from torchvision.models.resnet import resnet101
    model = resnet101(pretrained)

    ff = True if pretrained and fixed_feature else False
    return _ResNet(model, ff)

def resnet152(pretrained=False, fixed_feature=True):
    """ "ResNet-152 model from torchvision's resnet model.

    :param pretrained: if true, return a model pretrained on ImageNet
    :param fixed_feature: if true and pretrained is true, model features are fixed while training.
    """
    from torchvision.models.resnet import resnet152
    model = resnet152(pretrained)

    ff = True if pretrained and fixed_feature else False
    return _ResNet(model, ff)




"""
CheckpointHandler.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""

import torch

class CheckpointHandler:

    def store_var(self, var_name, value, exist_fail=False):
        if exist_fail is True and hasattr(self, var_name):
            raise Exception("var_name='{}' already exists".format(var_name))
        else:
            setattr(self, var_name, value)

    def get_var(self, var_name):
        if hasattr(self, var_name):
            value = getattr(self, var_name)
            return value
        else:
            return False

    def save_checkpoint(self, checkpoint_path, model, optimizer=None):
        if type(model) == torch.nn.DataParallel:
            # converting a DataParallel model to be able load later without DataParallel
            self.model_state_dict = model.module.state_dict()
        else:
            self.model_state_dict = model.state_dict()

        if optimizer:
            self.optimizer_state_dict = optimizer.state_dict()

        torch.save(self, checkpoint_path)

    @staticmethod
    def load_checkpoint(checkpoint_path, map_location='cpu'):
        checkpoint = torch.load(checkpoint_path, map_location=map_location)
        return checkpoint


"""
Validation functions. Follow functions are able to get scalar values for the validation.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import numpy as np

def label_accuracy_score(label_trues, label_preds, n_class):
    """
    :param label_trues:
    :param label_preds:
    :param n_class:
    :return: accuracy score and evaluation results
    		(overall accuracy, mean accuracy, mean IoU, fwavacc)
    """

    hist = np.zeros((n_class, n_class))
    for lt, lp in zip(label_trues, label_preds):
        hist += _fast_hist(lt.flatten(), lp.flatten(), n_class)
    acc = np.diag(hist).sum() / hist.sum()
    with np.errstate(divide='ignore', invalid='ignore'):
        acc_cls = np.diag(hist) / hist.sum(axis=1)
    acc_cls = np.nanmean(acc_cls)
    with np.errstate(divide='ignore', invalid='ignore'):
        iu = np.diag(hist) / (
            (hist.sum(axis=1) + hist.sum(axis=0) - np.diag(hist)).astype(np.float32)
        )

    mean_iou = np.nanmean(iu)
    freq = hist.sum(axis=1) / hist.sum()
    fwavacc = (freq[freq > 0] * iu[freq > 0]).sum()
    return acc, acc_cls, mean_iou, fwavacc

def _fast_hist(label_true, label_pred, n_class):
    mask = (label_true >= 0) & (label_true < n_class)
    hist = np.bincount(
        n_class * label_true[mask].astype(int) +
        label_pred[mask], minlength=n_class ** 2).reshape(n_class, n_class)
    return hist



"""
base class for pre-trained model.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division
from torch import nn
from dataclasses import dataclass

@dataclass
class CopyFeatureInfo:
	index: int
	out_channels: int

class SqueezeExtractor(nn.Module):
	def __init__(self, model, features, fixed_feature=True):
		super(SqueezeExtractor, self).__init__()
		self.model = model
		self.features = features
		if fixed_feature:
			for param in self.features.parameters():
				param.requires_grad = False

	def get_copy_feature_info(self):
		"""
		Get [CopyFeatureInfo] when sampling such as maxpooling or conv2d which has the 2x2 stride.
		:return: list. [CopyFeatureInfo]
		"""
		raise NotImplementedError()

	def _get_last_conv2d_out_channels(self, features):
		for idx, m in reversed(list(enumerate(features.modules()))):
			if isinstance(m, nn.Conv2d):
				return int(m.out_channels)
		assert False


"""
Mobilenet model Customized from Torchvision.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division
from .squeeze_extractor import *
from torch import nn

class _Mobilenet(SqueezeExtractor):
	def __init__(self, model, features, fixed_feature=True):
		layer = []
		layers = []
		self.zip_factor = [2, 4, 7, 11, 14, 17, 18]
		layer, layers = self._get_layers(features, layer, layers, 0)
		layers = nn.ModuleList(layers)
		super(_Mobilenet, self).__init__(model, layers, fixed_feature)

	def _get_layers(self, features, layer, layers, zip_cnt):
		from torchvision.models.mobilenet import InvertedResidual, ConvBNReLU

		for feature in features.children():
			if isinstance(feature, nn.Sequential) or\
					isinstance(feature, ConvBNReLU):
				layer, layers = self._get_layers(feature, layer, layers, zip_cnt)
				if zip_cnt == 0 and isinstance(feature, ConvBNReLU):
					layers += [nn.Sequential(*layer)]
					layer.clear()
			if isinstance(feature, InvertedResidual):
				layer, layers = self._get_layers(feature, layer, layers, zip_cnt)
				zip_cnt += 1
				if zip_cnt in self.zip_factor:
					layers += [nn.Sequential(*layer)]
					layer.clear()

			if len(list(feature.children())) == 0:
				layer += [feature]

		return layer, layers
	def get_copy_feature_info(self):
		lst_copy_feature_info = []
		channel = 0
		for i in range(len(self.features)):
			feature = self.features[i]
			if isinstance(feature, nn.MaxPool2d):
				lst_copy_feature_info.append(CopyFeatureInfo(i, channel))
			for idx, m in enumerate(feature.modules()):
				if isinstance(m, nn.Conv2d) and m.stride == (2, 2):
					channel = self._get_last_conv2d_out_channels(feature)
					lst_copy_feature_info.append(CopyFeatureInfo(i, channel))
					break

		return lst_copy_feature_info


def mobilenet(pretrained=False, fixed_feature=True):
	""" Mobile-net V2 model from torchvision's resnet model.

	:param pretrained: if true, return a model pretrained on ImageNet
	:param fixed_feature: if true and pretrained is true, model features are fixed while training.
	"""
	from torchvision.models.mobilenet import mobilenet_v2
	model = mobilenet_v2(pretrained)
	features = model.features[:-1]

	ff = True if pretrained and fixed_feature else False
	return _Mobilenet(model, features, ff)



"""
The transform method for the SegmentationDataset

Library:	Tensowflow 2.2.0, pyTorch 1.5.1, OpenCV-Python 4.1.1.26
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, print_function, division

import numpy as np
import torch
import cv2

class Rescale(object):
    """Rescale the image in a sample to a given size.

    Args:
        output_size (tuple or int): Desired output size. If tuple, output is
            matched to output_size. If int, smaller of image edges is matched
            to output_size keeping aspect ratio the same.
    """

    def __init__(self, output_size):
        assert isinstance(output_size, (int, tuple))
        self.output_size = output_size

    def __call__(self, sample):
        image, labeled = sample['image'], sample['labeled']

        h, w = image.shape[:2]
        if isinstance(self.output_size, int):
            if h > w:
                new_h, new_w = self.output_size * h / w, self.output_size
            else:
                new_h, new_w = self.output_size, self.output_size * w / h
        else:
            new_h, new_w = self.output_size

        new_h, new_w = int(new_h), int(new_w)

        img = cv2.resize(image, (new_w, new_h), interpolation=cv2.INTER_NEAREST)
        lbl = cv2.resize(labeled, (new_w, new_h), interpolation=cv2.INTER_NEAREST)

        return {'image': img, 'labeled': lbl}

class RandomHorizontalFlip(torch.nn.Module):
    """Horizontally flip the given image randomly with a given probability.
    The image can be a PIL Image or a torch Tensor, in which case it is expected
    to have [..., H, W] shape, where ... means an arbitrary number of leading
    dimensions

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def __call__(self, sample):
        image, labeled = sample['image'], sample['labeled']

        if torch.rand(1) < self.p:
            image = cv2.flip(image, 1)
            labeled = cv2.flip(labeled, 1)

        return {'image': image, 'labeled': labeled}

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)

class MakeSegmentationArray(object):
    """Make segmentation array from the annotation image"""

    def __init__(self, n_classes):
        assert isinstance(n_classes, int)

        self.n_classes = n_classes

    def __call__(self, sample):
        annotation = sample['annotation']
        assert annotation.dtype != int

        h, w = annotation.shape[:2]

        seg_labels = np.zeros((self.n_classes, h, w), dtype=annotation.dtype)

        for label in range(self.n_classes):
            seg_labels[label, :, :] = (annotation == label)

        return {'image': sample['image'], 'annotation': seg_labels}

class ToTensor(object):
    """Convert ndarrays in sample to Tensors."""

    def __call__(self, sample):
        image, lbl = sample['image'], sample['labeled']

        # swap color axis because
        # numpy image: H x W x C
        # torch image: C X H X W
        image = image.transpose((2, 0, 1))
        return {'image': torch.from_numpy(image).float(),
                'annotation': torch.from_numpy(lbl).long()}


"""
Unet class.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import torch
from ..encoders.squeeze_extractor import *

class UnetWithEncoder(torch.nn.Module):

    def __init__(self, n_classes, pretrained_model: SqueezeExtractor, batch_norm=True):
        super(UnetWithEncoder, self).__init__()
        self.copy_feature_info = pretrained_model.get_copy_feature_info()
        self.features = pretrained_model.features

        self.up_layer0 = self._make_up_layer(-1, batch_norm)
        self.up_layer1 = self._make_up_layer(-2, batch_norm)
        self.up_layer2 = self._make_up_layer(-3, batch_norm)
        self.up_layer3 = self._make_up_layer(-4, batch_norm)

        self.up_sampling0 = self._make_up_sampling(-1)
        self.up_sampling1 = self._make_up_sampling(-2)
        self.up_sampling2 = self._make_up_sampling(-3)
        self.up_sampling3 = self._make_up_sampling(-4)

        #find out_channels of the top layer and define classifier
        for f in reversed(self.up_layer3):
            if isinstance(f, nn.Conv2d):
                channels = f.out_channels
                break

        uplayer4 = []
        uplayer4 += [nn.Conv2d(channels, channels, kernel_size=3, padding=1)]
        if batch_norm:
            uplayer4 += [nn.BatchNorm2d(channels)]
        uplayer4 += [nn.ReLU(inplace=True)]
        self.up_layer4 = nn.Sequential(*uplayer4)

        self.up_sampling4 = nn.ConvTranspose2d(channels, channels, kernel_size=4,
                                              stride=2, bias=False)
        self.classifier = nn.Sequential(nn.Conv2d(channels, n_classes, kernel_size=1), nn.ReLU(inplace=True))
        self._initialize_weights()

    def _get_last_out_channels(self, features):
        for idx, m in reversed(list(enumerate(features.modules()))):
            if isinstance(m, nn.Conv2d):
                return m.out_channels
        return 0


    def _make_up_sampling(self, cfi_idx):
        if cfi_idx == -1:
            in_channels = self._get_last_out_channels(self.features)
        else:
            in_channels = self.copy_feature_info[cfi_idx + 1].out_channels

        out_channels = self.copy_feature_info[cfi_idx].out_channels
        return nn.ConvTranspose2d(in_channels, out_channels, kernel_size=4,
                                  stride=2, bias=False)

    def _make_up_layer(self, cfi_idx, batch_norm):
        idx = self.copy_feature_info[cfi_idx].index
        for k in reversed(range(0, idx)):
            f = self.features[k]
            channels = self._get_last_out_channels(f)

            if channels == 0:
                continue

            out_channels = self.copy_feature_info[cfi_idx].out_channels
            in_channels = out_channels + channels  # for concatenation.

            layer = []
            layer += [nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)]
            if batch_norm:
                layer += [nn.BatchNorm2d(out_channels)]
            layer += [nn.ReLU(inplace=True)]

            return nn.Sequential(*layer)

        assert False

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        copy_out = []
        o = x
        cpi = self.copy_feature_info[-4:]
        copy_idx = 0

        for i in range(len(self.features)):
            o = self.features[i](o)
            if i == cpi[copy_idx].index - 1:
                copy_out.append(o)
                if copy_idx + 1 < len(cpi):
                    copy_idx += 1

        o = self.up_sampling0(o)
        o = o[:, :, 1:1 + copy_out[3].size()[2], 1:1 + copy_out[3].size()[3]]
        o = torch.cat([o, copy_out[3]], dim=1)
        o = self.up_layer0(o)

        o = self.up_sampling1(o)
        o = o[:, :, 1:1 + copy_out[2].size()[2], 1:1 + copy_out[2].size()[3]]
        o = torch.cat([o, copy_out[2]], dim=1)
        o = self.up_layer1(o)

        o = self.up_sampling2(o)
        o = o[:, :, 1:1 + copy_out[1].size()[2], 1:1 + copy_out[1].size()[3]]
        o = torch.cat([o, copy_out[1]], dim=1)
        o = self.up_layer2(o)

        o = self.up_sampling3(o)
        o = o[:, :, 1:1 + copy_out[0].size()[2], 1:1 + copy_out[0].size()[3]]
        o = torch.cat([o, copy_out[0]], dim=1)
        o = self.up_layer3(o)

        o = self.up_sampling4(o)
        cx = int((o.shape[3] - x.shape[3]) / 2)
        cy = int((o.shape[2] - x.shape[2]) / 2)
        o = o[:, :, cy:cy + x.shape[2], cx:cx + x.shape[3]]
        o = self.up_layer4(o)
        o = self.classifier(o)

        return o

cfgs = {
    'A': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 1024, 1024, 'U', 512, 512,
           'U', 256, 256, 'U', 128, 128, 'U', 64, 64]
}

class Unet(torch.nn.Module):

    def __init__(self, n_classes, cfg, batch_norm=True):
        super(Unet, self).__init__()
        self.features = self._make_layers(cfg, batch_norm)
        self.classifier = nn.Sequential(nn.Conv2d(cfg[-1], n_classes, kernel_size=1), nn.Sigmoid())

        self._initialize_weights()


    def _make_layers(self, cfg, batch_norm):
        layers = []
        in_channels = 3
        for v in cfg:
            if v == 'M':
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            elif v == 'U':
                layers += [nn.ConvTranspose2d(in_channels, int(in_channels / 2), kernel_size=4,
                                               stride=2, bias=False)]
            else:
                conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
                if batch_norm:
                    layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
                else:
                    layers += [conv2d, nn.ReLU(inplace=True)]
                in_channels = v
        return nn.Sequential(*layers)



    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        o = x
        size_features = len(self.features)
        copys = []
        for i in range(size_features):
            o = self.features[i](o)
            if isinstance(self.features[i], nn.ConvTranspose2d):
                copy = copys.pop()
                o = o[:, :, 1:1 + copy.size()[2], 1:1 + copy.size()[3]]
                o = torch.cat([o, copy], dim=1)
            if i + 1 >= size_features:
                continue
            if isinstance(self.features[i+1], nn.MaxPool2d):
                copys += [o]

        cx = int((o.shape[3] - x.shape[3]) / 2)
        cy = int((o.shape[2] - x.shape[2]) / 2)
        o = o[:, :, cy:cy + x.shape[2], cx:cx + x.shape[3]]
        o = self.classifier(o)

        return o

from ..encoders.vgg import *
from ..encoders.resnet import *
from ..encoders.mobilenet import *

def unet_vgg11(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_11(batch_norm, pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, vgg, batch_norm)
def unet_vgg13(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_13(batch_norm, pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, vgg, batch_norm)
def unet_vgg16(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_16(batch_norm, pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, vgg, batch_norm)
def unet_vgg19(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_19(batch_norm, pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, vgg, batch_norm)

def unet_resnet18(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet18(pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, resnet, batch_norm)
def unet_resnet34(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet34(pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, resnet, batch_norm)
def unet_resnet50(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet50(pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, resnet, batch_norm)
def unet_resnet101(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet101(pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, resnet, batch_norm)
def unet_resnet152(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet152(pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, resnet, batch_norm)

def unet_mobilenet_v2(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    mobile_net = mobilenet(pretrained, fixed_feature)
    return UnetWithEncoder(n_classes, mobile_net, batch_norm)

def unet(n_classes, batch_size):
    batch_norm = False if batch_size == 1 else True
    return Unet(n_classes, cfgs['A'], batch_norm)




import torch
from torchvision import transforms

from pytorch.segmentation.data_loader.segmentation_dataset import SegmentationDataset
from pytorch.segmentation.data_loader.transform import Rescale, ToTensor
from pytorch.segmentation.trainer import Trainer
from pytorch.segmentation.predict import *
from pytorch.segmentation.models import all_models
from pytorch.util.logger import Logger

# train_images = r'D:/datasets/cityspaces_full/images/train'
# test_images = r'D:/datasets/cityspaces_full/images/test'
# train_labled = r'D:/datasets/cityspaces_full/labeled/train'
# test_labeled = r'D:/datasets/cityspaces_full/labeled/test'

train_images = r'test/dataset/cityspaces/images/train'
test_images = r'test/dataset/cityspaces/images/test'
train_labled = r'test/dataset/cityspaces/labeled/train'
test_labeled = r'test/dataset/cityspaces/labeled/test'

if __name__ == '__main__':
    model_name = "fcn16_resnet18"
    device = 'cuda'
    batch_size = 8
    n_classes = 34
    input_axis_minimum_size = 80
    fixed_feature = False
    pretrained = False
    num_epochs = 1
    check_point_stride = 1

    logger = Logger(model_name=model_name, data_name='example')

    # Loader
    compose = transforms.Compose([
        Rescale(input_axis_minimum_size),
        ToTensor()
         ])

    train_datasets = SegmentationDataset(train_images, train_labled, n_classes, compose)
    train_loader = torch.utils.data.DataLoader(train_datasets, batch_size=batch_size, shuffle=True, drop_last=True)

    #test_datasets = SegmentationDataset(test_images, test_labeled, n_classes, compose)
    #test_loader = torch.utils.data.DataLoader(test_datasets, batch_size=batch_size, shuffle=True, drop_last=True)
    test_loader = None

    # Model
    model = all_models.model_from_name[model_name](n_classes, batch_size,
                                                   pretrained=pretrained,
                                                   fixed_feature=fixed_feature)
    #Optimizers

    if pretrained and fixed_feature: #fine tunning
        params_to_update = model.parameters()
        print("Params to learn:")
        params_to_update = []
        for name, param in model.named_parameters():
            if param.requires_grad == True:
                params_to_update.append(param)
                print("\t", name)
        optimizer = torch.optim.Adadelta(params_to_update)
    else:
        optimizer = torch.optim.Adadelta(model.parameters())

    #logger.load_model(model, 'epoch_280')
    model.to(device)
    #Train
    scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=10, gamma=0.1)
    trainer = Trainer(model, optimizer, logger, num_epochs,
                      train_loader, test_loader,
                      #epoch=281,
                      check_point_epoch_stride=check_point_stride)
    trainer.train()

    # Get result
    predict(model, r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\input3.png',
              r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\output3.png')
    # predict(model, r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\input4.png',
    #         r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\output4.png')
    # predict(model, r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\input3.png',
    #         r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\output3.png')
    #
    # convert_seg_gray_to_color(r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\ground_truth3.png', n_classes,
    #                           output_path='C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\c_ground_truth3.png')
    # convert_seg_gray_to_color(r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\ground_truth4.png', n_classes,
    #                           output_path='C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\c_ground_truth4.png')
    # convert_seg_gray_to_color(r'C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\ground_truth5.png', n_classes,
    #                           output_path='C:\Dropbox\Dropbox\python\pytorch\docs\exam_result\c_ground_truth5.png')







from __future__ import print_function

from setuptools import find_packages, setup


def get_long_description():
    with open('README.md') as f:
        long_description = f.read()

    try:
        import github2pypi

        return github2pypi.replace_url(
            slug='IanTaehoonYoo/semantic-segmentation-pytorch', content=long_description
        )
    except Exception:
        return long_description

setup(name="seg_torch",
      version="0.1.6",
      description="Semantic Segmentation with Pytorch",
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      author="Ian Yoo",
      author_email='thyoostar@gmail.com',
      platforms=["any"],
      license="MIT",
      url="https://github.com/IanTaehoonYoo/semantic-segmentation-pytorch",
      packages=find_packages(exclude=["segmentation/test/dataset"]),
      install_requires=[
            "torch>=1.5.0",
            "torchvision>=0.5.0",
            "tensorboardX>=2.0"
            "opencv-python",
            "tqdm"],
      extras_require={
            "tensorflow": ["tensorflow"], #this is to provide backbone models.
      },
      classifiers=['License :: OSI Approved :: MIT License']
)


"""
FCN32 class.

Library:	Tensowflow 2.2.0, pyTorch 1.5.1
Author:		Ian Yoo
Email:		thyoostar@gmail.com
"""
from __future__ import absolute_import, division, print_function

import torch
from ..encoders.squeeze_extractor import *

class FCN32(torch.nn.Module):

    def __init__(self, n_classes, pretrained_model: SqueezeExtractor):
        super(FCN32, self).__init__()
        self.pretrained_model = pretrained_model
        self.features = self.pretrained_model.features
        self.upsampling32 = nn.ConvTranspose2d(n_classes, n_classes, kernel_size=64,
                                               stride=32, bias=False)

        for m in self.features.modules():
            if isinstance(m, nn.Conv2d):
                channels = m.out_channels

        self.classifier = nn.Sequential(nn.Conv2d(channels, n_classes, kernel_size=1), nn.Sigmoid())
        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        o = x
        for feature in self.features:
            o = feature(o)
        o = self.classifier(o)
        o = self.upsampling32(o)
        cx = int((o.shape[3] - x.shape[3]) / 2)
        cy = int((o.shape[2] - x.shape[2]) / 2)
        o = o[:, :, cy:cy + x.shape[2], cx:cx + x.shape[3]]

        return o

from ..encoders.vgg import *
from ..encoders.resnet import *
from ..encoders.mobilenet import *

def fcn32_vgg11(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_11(batch_norm, pretrained, fixed_feature)
    return FCN32(n_classes, vgg)
def fcn32_vgg13(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_13(batch_norm, pretrained, fixed_feature)
    return FCN32(n_classes, vgg)
def fcn32_vgg16(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_16(batch_norm, pretrained, fixed_feature)
    return FCN32(n_classes, vgg)
def fcn32_vgg19(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    vgg = vgg_19(batch_norm, pretrained, fixed_feature)
    return FCN32(n_classes, vgg)

def fcn32_resnet18(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet18(pretrained, fixed_feature)
    return FCN32(n_classes, resnet)
def fcn32_resnet34(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet34(pretrained, fixed_feature)
    return FCN32(n_classes, resnet)
def fcn32_resnet50(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet50(pretrained, fixed_feature)
    return FCN32(n_classes, resnet)
def fcn32_resnet101(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet101(pretrained, fixed_feature)
    return FCN32(n_classes, resnet)
def fcn32_resnet152(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    resnet = resnet152(pretrained, fixed_feature)
    return FCN32(n_classes, resnet)

def fcn32_mobilenet_v2(n_classes, batch_size, pretrained=False, fixed_feature=True):
    batch_norm = False if batch_size == 1 else True
    mobile_net = mobilenet(pretrained, fixed_feature)
    return FCN32(n_classes, mobile_net)
